#!/usr/bin/env python

import hashlib
import json
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
import yaml

logging.basicConfig(level=logging.DEBUG, format='%(levelname)s %(message)s')

URL_REGEX = re.compile(r'^http(s?)\:\/\/')
RPM_REGEX = re.compile(r'\.rpm$')
FP_REGEX = re.compile(r'\.fp$')
CHECKSUMS = ['md5', 'sha1', 'sha256', 'sha384', 'sha512']

class Repo:
    def __init__(self):
        """
        Class constructor
        Initiate all required stuff
        """
        self._metadata = []

    @property
    def __json__(self):
        """
        Return repo data in JSON format
        """
        return json.dumps(self._metadata)

    @property
    def __yaml__(self):
        """
        Return repo data in yaml format
        """
        return yaml.dump(self._metadata, default_flow_style=False)

    def _checksums(self, file):
        logging.info('Generating checksums for %s' % file)
        out = {}
        for checksum in CHECKSUMS:
            fp = open(file, 'rb')
            h = hashlib.new(checksum)
            h.update(fp.read())
            fp.close()

            out[checksum] = h.hexdigest()
            logging.debug('%s: %s' % (checksum, out[checksum]))

        logging.debug('Writing checksums to %s.sha.txt' % file)
        fp = open('%s.sha.txt' % file, 'w')
        fp.write(yaml.dump(out, default_flow_style=False))
        fp.close()

        return out

    def _cmd(self, command):
        logging.debug('Executing: %s' % command)
        p = subprocess.Popen(command,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        return p.communicate()

    @property
    def _files(self):
        (files, err) = self._cmd("find | sed 's~^\.\/~~'")

        out = []
        for file in files.split("\n"):
            if re.search(FP_REGEX, file) or re.search(RPM_REGEX, file):
                out.append(file)

        return out

    def _flush(self):
        fp = open('metadata.yaml', 'w')
        fp.write(self.__yaml__)
        fp.close()

        fp = open('metadata.json', 'w')
        fp.write(self.__json__)
        fp.close()

    def _index(self):
        for file in self._files:
            searchdir = self._unpack(file)

            (out, err) = self._cmd('find %s -maxdepth 2 -name metadata.yaml' % searchdir)
            metadata = out.strip()

            logging.debug('Found metadata.yaml: %s' % metadata)
            self._metadata.append(
                self._single_metadata(metadata, file))

            logging.info('Cleaning up %s' % searchdir)
            self._cmd('rm -rf %s' % searchdir)
        self._flush()

    def _single_metadata(self, metadatayaml, file):
        fp = open(metadatayaml, 'r')
        data = yaml.load(fp)

        docurl = 'docs/%s/%s/%s/%s-%s.pdf' % (
            data['name'][0],
            data['name'][1],
            data['name'],
            data['name'],
            data['version']
        )

        if os.path.exists(docurl):
            data['doc'] = docurl
            data['docs-checksum'] = self._checksums(docurl)
        data['binary'] = file
        data['binary-checksum'] = self._checksums(file)

        fp.close()
        return data

    def _unpack(self, file):
        logging.info('Unpacking %s' % file)
        tmpdir = tempfile.mkdtemp(suffix='.tmp', prefix='plugins_')
        if re.search(FP_REGEX, file):
            (out, err) = self._cmd('tar xfz %s -C %s' % (file, tmpdir))
            searchdir = tmpdir
        elif re.search(RPM_REGEX, file):
            (out, err) = self._cmd('cd "%s" ; rpm2cpio %s | cpio -idmv' % (tmpdir, os.path.abspath(file)))
            searchdir = '%s/var/www/nailgun/plugins' % tmpdir
        else:
            pass

        return searchdir

    def add(self, plugin, doc):
        """
        Add plugin to repository
        """
        searchdir = self._unpack(plugin)
        (out, err) = self._cmd('find %s -maxdepth 2 -name metadata.yaml' % searchdir)
        metadata = out.strip()

        data = yaml.load(open(metadata, 'r').read())

        if re.search(FP_REGEX, plugin):
            pluginpath = 'repository/%s/%s/%s/%s-%s.fp' % (
                data['name'][0],
                data['name'][1],
                data['name'],
                data['name'],
                data['version']
            )
        elif re.search(RPM_REGEX, plugin):
            pluginpath = 'repository/%s/%s/%s/%s-%s.noarch.rpm' % (
                data['name'][0],
                data['name'][1],
                data['name'],
                data['name'],
                data['version']
            )
        else:
            raise

        docpath = 'docs/%s/%s/%s/%s-%s.pdf' % (
            data['name'][0],
            data['name'][1],
            data['name'],
            data['name'],
            data['version']
        )

        try:
            os.makedirs(os.path.dirname(pluginpath))
        except OSError as e:
            if str(e).startswith('[Errno 17] File exists:'):
                pass
            else:
                raise

        try:
            os.makedirs(os.path.dirname(docpath))
        except OSError as e:
            if str(e).startswith('[Errno 17] File exists:'):
                pass
            else:
                raise

        shutil.copy(plugin, pluginpath)
        shutil.copy(doc, docpath)

        self.index()

    def help(self):
        print """
fprepo COMMAND [OPTIONS]
    index           index repository and generate metadata.yaml
    init            init empty plugin repository
    add PLUGIN DOC  add plugin to repository and reindex
    del PLUGIN      remove plugin from repository and reindex
        """

        sys.exit(1)

    def index(self):
        self._index()

    def init(self):
        """
        Init empty plugins repository
        """
        os.mkdir('repository')
        os.mkdir('docs')

        self._flush()

    def remove(self, name):
        """
        Remove plugin from repository
        """
        logging.info('Trying to remove plugin %s' % name)
        logging.debug('Reading metadata.yaml')
        fp = open('metadata.yaml', 'r')
        data = yaml.load(fp)
        fp.close()

        c = 0
        for plugin in data:
            if plugin['name'] == name:
                c = c + 1
                logging.info('Removing %s' % plugin['name'])

                files = [
                    plugin['binary'],
                    '%s.sha.txt' % plugin['binary'],
                    plugin['doc'],
                    '%s.sha.txt' % plugin['doc'],
                ]
                for file in files:
                    logging.debug('Unlinking %s' % file)
                    os.unlink(file)

                os.removedirs(os.path.dirname(plugin['binary']))
                os.removedirs(os.path.dirname(plugin['doc']))

        if c == 0:
            logging.error('No plugin named %s found in metadata.yaml. Pobably you need index repository first.' %
                self.args[1])
            sys.exit(1)
        self.index()
